"use strict";

const Diagnostics = function() {

  /**
   * @param {String} serial_number
   * @return {File}
   */
  let getBasicInfoFile = function (serial_number)
  {
    let data = [
      [ 'Local date', (new Date()).toLocaleString() ],
      [ 'UTC date', (new Date()).toUTCString() ],
      [ 'User Agent', window.navigator.userAgent ],
      [ 'Current URL', document.location.href ],
      [ 'Referrer', document.referrer ],
      [ 'Serial number', serial_number],
    ];
    let data_string = data.map( item => `${item[0]}: ${item[1]}`).join("\n");
    return new File([data_string], 'basic.txt', {type: 'text/plain'});
  };

  /**
   * @param {String} config_arr
   * @param {String} config_arr_tmp
   * @return {File}
   */
  let getConfigInfoFile = function (config_arr, config_arr_tmp)
  {
    let data = [
      [ 'Config Array', config_arr],
      [ 'Config Array Tmp', config_arr_tmp],
    ];
    let data_string = data.map( item => `${item[0]}: ${item[1]}`).join("\n\n");
    return new File([data_string], 'config.txt', {type: 'text/plain'});
  };

  /**
   * @param {string} mode [text, zip, zip_with_image]
   * @return {Promise<Text>}
   */
  let getServerReport = async function(mode = 'text') {
    try {
      const url = '/axis-cgi/admin/serverreport.cgi?mode=' + mode;
      const response = await fetch(url);
      if (!response.ok) {
        throw new Error('HTTP error, status = ' + response.status);
      }

      return response.blob();
    } catch (e) {
      throw new Error(`Failed to fetch server report. ${e.message}`);
    }
  };

  /**
   * @return {Promise<File>}
   */
  let getServerReportFile = async function()
  {
    const report = await getServerReport('zip');
    return new File([report], 'server_report.zip', { type: "application/zip"});
  };

  /**
   * @param report_msg
   * @param {String} serial_number
   * @param {String} config_arr
   * @param {String} config_arr_tmp
   * @return {Promise<Blob>}
   */
  let createReport = async function (report_msg, serial_number, config_arr, config_arr_tmp)
  {
    const attachments = [];

    report_msg.updateText('Fetching Server report from camera');
    try {
      const server_report = await getServerReportFile();
      attachments.push({name: 'server_report', data: server_report});
    } catch (err) {
      // failure to get server report should not stop building diagnostics
      report_msg.hide();
      showMessage('warning', 'Failed to fetch server report from camera.');
      console.warn('Failed to fetch server report from camera.', err);
    }

    report_msg.updateText('Preparing report file');

    const config_file = getConfigInfoFile(config_arr, config_arr_tmp);
    attachments.push({name: 'config', data: config_file});

    const basic_file = getBasicInfoFile(serial_number);
    attachments.push({name: 'basic', data: basic_file});

    const console_file = await getConsoleLogFile();
    attachments.push({name: 'console_log', data: console_file});

    return await getReport(attachments);
  };

  /**
   * @return {Promise<File>}
   */
  let getConsoleLogFile = async function()
  {
    const logs = await jsLogger.getLogs();
    const logs_json = JSON.stringify(logs, null, 2);
    return new File([logs_json], 'console_log.txt', {type: 'text/plain'})
  };

  /**
   * @param {Array.<{name: String, data: Blob}>} attachments
   * @return {Promise<Blob>}
   */
  let getReport = async function(attachments = []) {

    let data = new FormData();
    attachments.forEach( item => data.set(item.name, item.data));

    let request = new Request("/local/camoverlay/api/report.cgi", {
      method: 'POST',
      body: data,
      cache: "no-cache",
    });

    let response;
    try {
      response = await fetch(request);
    } catch (e) {
      throw new Error(`Failed to fetch report. ${e.message}`);
    }

    const content_type = response.headers.get('Content-Type');
    if (content_type === 'application/zip') {
      return response.blob();
    } else if (content_type === 'application/json') {
      const json = response.json();
      throw new Error(json.message || json.status || 'Invalid response from camera.');
    } else {
      throw new Error(`Invalid response content type '${content_type}'.`)
    }
  };

  /**
   * @param report_msg
   * @param {string} config_arr
   * @param {string} config_arr_tmp
   * @return {Promise<void>}
   */
  let downloadReport = async function(report_msg, config_arr, config_arr_tmp) {

    let serial_number_param = await getParameter('Properties.System.SerialNumber,CamOverlay.').catch(err => { throw new Error(err);});
    let serial_number = serial_number_param['properties.system.serialnumber'];
    let report_blob = await createReport(report_msg, serial_number, config_arr, config_arr_tmp);

    let a = document.createElement("a");
    let url = window.URL.createObjectURL(report_blob);
    a.style.display = "none";
    a.href = url;
    a.download = `camoverlay_report_${serial_number}.zip`;

    document.body.appendChild(a);
    a.click();
    window.URL.revokeObjectURL(url);
    document.body.removeChild(a);
  };

  return Object.freeze({
    downloadReport
  });
};