function FileManager(app_name, options) {

  const fileManagerUrl = 'fileManager.html';
  const fileManagerModalId = 'fileManagerModal';
  let fileList = [];

  /**
   * Load HTML file to container
   * @param html_container_id
   */
  let load = async function(html_container_id) {
    let fileManagerTemplate = await $.get(fileManagerUrl);
    $('#' + html_container_id).html(fileManagerTemplate);
  }

  /**
   * Open modal
   */
  let open = function() {
    $('#fileManagerModalTitle').text(options.modalTitle);
    $('#' + fileManagerModalId).modal({show: true});
    options.showGalleryAlert();
  }

  let init = async function() {
    await load('fileManager');
    await createFileList();
    await initAllowedStorage();
    
    $('#fileStorageLabel').attr('title', options.messages.labelFileInputTooltip).tooltip();
    $('#fileStorage').change(storageChanged);
    $('#uploadedFile').change(validateForm);
    $('.fileUpload').click(uploadClicked);
  }

  /**
   * Get list of images/files
   * @returns {Promise<HTMLElement|[]|[number, number, [], string, string]>}
   */
  let getFileList = async function() {
    let fileListResponse = null;
    try {
      fileListResponse = await $.get(options.api.listUrl);
    } catch (e) {
      throw new Error("Error while getting list of files.");
    }
    return fileListResponse.list.sort((a, b) => a.name < b.name ? -1 : (a.name > b.name ? 1 : 0));
  }

  /**
   * Get nice and modified tooltip text with img path
   * @param {string} path
   * @returns {string}
   */
  let getNicePathTooltip = function(path) {
    let tooltip = escapeHtml(path);
    if (tooltip.indexOf('file://') !== -1) {
      tooltip = tooltip.substring(7);
    }
    return tooltip;
  }

  /**
   * Convert storage string to storage nice name
   * @type {{ftp: string, samba: string, SD0: string, url: string, flash: string}}
   */
  const convertStorageTypeToNiceName = {
    flash: 'Internal',
    SD0: 'SD Card',
    ftp: 'FTP',
    samba: 'MSNS',
    url: 'URL'
  };

  /**
   * Cut longer name than 40 chars and end with ...filetype
   * @param {string} name
   * @returns {string}
   */
  let getNiceName = function(name) {
    const nameMaxLength = 40;
    if (name.length > nameMaxLength) {
      let fileType = name.split(".")[1];
      name = name.substring(0, nameMaxLength) + "...." + fileType;
    }
    return name;
  }

  /**
   * Create html table row for file list
   * @param {number} itemIndex
   * @param {object} itemData
   * @returns {string}
   */
  let createFileListRow = function(itemIndex, itemData) {
    let tableRow = '<tr id="imgId' + itemIndex + '" data="' + itemIndex + '" data-toggle="tooltip" title="' + getNicePathTooltip(itemData.path) + '">';
    tableRow +=   '<td class="text-center"><span class="font-weight-normal badge badge-secondary">' + convertStorageTypeToNiceName[itemData.storage] + '</span></td>';
    tableRow +=   '<td class="text-left">' + getNiceName(itemData.name) + '</td>';
    tableRow +=   '<td><button data="' + itemIndex + '" class="btn btn-sm btn-link deleteFile"><span class="fa fa-trash text-primary"></span></button></td>';
    tableRow += '</tr>';
    return tableRow;
  }

  /**
   * Create file list
   * @returns {Promise<void>}
   */
  let createFileList = async function() {
    let fileListJqueryObject = $('#fileList tbody');
    let html = [];

    fileList = await getFileList();

    $('#fileListContainer').addClass(options.listFileContainerWith);

    if (fileList.length === 0) { // Empty list
      const emptyRow = '<tr><td>' + options.messages.emptyRow + '</td></tr>';
      html.push(emptyRow);
    } else {
      fileList.forEach((item, index) => {
        let tableRow = createFileListRow(index, item);
        html.push(tableRow);
      });
    }

    fileListJqueryObject.empty().append(html.join(""));

    $('#fileList tr').click(imageClickedCallback);
    if (fileList.length > 0) {
      if ($('#fileList tr').length > 0) {
        $('#fileList tr')[0].click();
      }
    } else {
      $('#fileListImagePreview').empty();
    }

    options.onAfterCreateFileList();

    $('.deleteFile').click(deleteFileClickedCallback);
    $('.tooltip').remove();
    $('[data-toggle="tooltip"]').tooltip();
  }

  /**
   * Get allowed storage types
   * @returns {Promise<any>}
   */
  let getAllowedStorage = async function() {
    return $.get(options.api.getStorageUrl).catch(
      (jqXHR, textStatus, errorThrown) => {
        throw new Error(errorThrown || 'Network error');
      }
    );
  }

  /**
   * Init allowed storage types (creates options for select)
   * @returns {Promise<void>}
   */
  let initAllowedStorage = async function() {
    let fileStorageJqueryObject = $('#fileStorage');
    let storageResponse = null;
    try {
      storageResponse = await getAllowedStorage();
      if (storageResponse.code !== 200) {
        throw new Error("Error while getting storage data");
      }
    } catch (error) {
      throw new Error(error.message);
    }
    await storageResponse.list.forEach(item => fileStorageJqueryObject.append(createStoreOption(item)));
    storageChanged();
  }

  /**
   * Create HTML option element for storage select
   * @param {object<string, string>} item
   * @returns {HTMLOptionElement}
   */
  let createStoreOption = function(item) {
    let newOption = new Option(item.state, item.type, false, false);
    if (item.state.indexOf('non-writable') !== -1) {
      $(newOption).prop('disabled', true);
    }
    return newOption;
  }

  /**
   * Callback for delete btn
   * @param event
   * @returns {Promise<void>}
   */
  let deleteFileClickedCallback = async function(event) {
    let fileNumber = $(event.target).closest('tr').attr('data');
    let file = fileList[fileNumber];
    await $.get(options.api.removeUrl + '&storage=' + file.storage + '&path=' + encodeURIComponent(file.path) + '&name=' + encodeURIComponent(file.name));

    options.onAfterRemoveFile(file.path);

    createFileList();
  }

  const storageUrlSambaFtp = ['url', 'samba', 'ftp'];

  const alert = {
    failedToUpload: {
      text: 'Failed to upload file(s).',
      show: function() {
        let alertHTML = '<div id="failedToUpload" class="alert alert-danger alert-dismissible mb-2">' + this.text + '</div>';

        $('#alertBox')
          .empty()
          .append(alertHTML);

        setTimeout(function() {
          $('#failedToUpload').hide();
        }, 5000);
      }
    },
    removeAll: function() {
      $('#alertBox').empty();
    }
  };

  /**
   * Callback for upload btn
   * @returns {Promise<boolean>}
   */
  let uploadClicked = async function(e) {
    if (!validateForm()) {
      return false;
    }
    var forms = [];
    var storage = $('#fileStorage').val();
    if (storageUrlSambaFtp.includes(storage)) {
      var urlName = $('#urlName').val().trim();
      var url = $('#url').val().trim();
      if (storage === 'samba') {
        url = 'file:///var/spool/storage/NetworkShare/' + url;
      }
      let form = new FormData();
      form.append("target", storage);
      form.append("name", urlName);
      form.append("url", url);
      forms.push(form);
    } else {
      let formData = new FormData($('#imgForm')[0]);
      // If edge, safari or IE upload all images at once
      if (navigator.userAgent.indexOf('MSIE') !== -1 ||
        navigator.userAgent.indexOf('Trident') !== -1 ||  // IE 11+
        navigator.userAgent.indexOf('Edge') !== -1 ||
        /^((?!chrome|android).)*safari/i.test(navigator.userAgent)) { // Safari
        forms.push(formData);
      } else {
        // Find target
        for (var pair of formData.entries()) {
          if (pair[0] !== "target") {
            let formTmp = new FormData();
            formTmp.append("target", formData.get("target"));
            formTmp.append(pair[0], pair[1]);
            forms.push(formTmp);
          }
        }
      }
    }

    let ajaxList = [];
    for (var i = 0; i < forms.length; i++) {
      ajaxList.push(
        $.ajax({
          url: options.api.uploadUrl,
          data: forms[i],
          type: 'POST',
          contentType: false, // NEEDED, DON'T OMIT THIS (requires jQuery 1.6+)
          processData: false, // NEEDED, DON'T OMIT THIS
        })
      );
    }

    const fileTableTbodyJqueryObject = $('#' + fileManagerModalId + ' table tbody');
    fileTableTbodyJqueryObject.showSpinner('Uploading file');
    alert.removeAll();

    let failed = false;
    try {
      await Promise.all(ajaxList);
    } catch (e) {
      failed = true;
    }

    createFileList();
    fileTableTbodyJqueryObject.hideSpinner();

    if (failed) {
      alert.failedToUpload.show();
    }

    let forceToClose = $(e.target).hasClass('forceToClose');
    if (forceToClose && !failed) {
      $('#' + fileManagerModalId).modal('hide');
    }
  }

  /**
   * Validate form before post data
   * @returns {boolean}
   */
  let validateForm = function() {
    let storage = $('#fileStorage').val();
    if (storageUrlSambaFtp.includes(storage)) {
      // Name URL
      const urlName = $('#urlName').val();
      if (urlName.length === 0) {
        $('#urlNameFileGroup').addClass('has-error');
        $('#urlNameFileHelp').text(options.messages.urlName);
        return false;
      } else {
        $('#urlNameFileGroup').removeClass('has-error');
        $('#urlNameFileHelp').text('');
      }
      // URL
      const url = $('#url').val();
      const urlContainsProtocols = /(file|ftp|http|https):\/\//g.test(url);
      let urlFileHelpMsg = '';
      if (url.length === 0 || url.length > options.maxUrlLength || !urlContainsProtocols ) {
        $('#urlFileGroup').addClass('has-error');
        if (url.length === 0) {
          urlFileHelpMsg = options.messages.urlToShort;
        } else if (url.length > options.maxUrlLength) {
          urlFileHelpMsg = options.messages.urlToLong;
        } else if (!urlContainsProtocols) {
          urlFileHelpMsg = options.messages.urlFile;
        }
        $('#urlFileHelp').text(urlFileHelpMsg);
        return false;
      } else {
        $('#urlFileGroup').removeClass('has-error');
        $('#urlFileHelp').text(urlFileHelpMsg);
      }
    } else {
      // Upload file
      var inputFileArr = $('#uploadedFile')[0].files;
      var err = inputFileArr.length == 0;
      for (var i = 0; i < inputFileArr.length; i++) {
        let fileName = inputFileArr[i].name.toLowerCase();
        fileName = fileName.split(".");
        let fileType = fileName[fileName.length-1];
        if(!options.filesAllowedTypes.includes(fileType)) {
          err = true;
          break;
        }
      }
      if (err) {
        $('#uploadedFileGroup').addClass('has-error');
        $('#uploadedFileHelp').text(options.messages.filesAllowed);
        return false;
      } else {
        $('#uploadedFileGroup').removeClass('has-error');
        $('#uploadedFileHelp').text('');
      }
    }
    return true;
  }

  /**
   * Updates image preview (only for imageManager)
   * @param event
   */
  let imageClickedCallback = async function(event) {
    let target = event.target;
    if ($(target).hasClass('deleteFile')) {
      return;
    }
    if (fileList.length === 0) {
      return;
    }
    let fileId = $(target).closest('tr').attr('id');
    let fileNumber = $(target).closest('tr').attr('data');
    let encodePath = encodeURIComponent(fileList[fileNumber].path);
    let imagePreviewSrc = String($('#imageManagerPreview').attr('src'));
    // Replace image preview only if path is changed
    if (imagePreviewSrc.indexOf(encodePath) === -1) {
      // Set img and insert to html
      let fileSrc = '/local/' + app_name + '/api/image.cgi?path=' + encodePath + '&_t=' + Date.now();
      $('#fileListImagePreview').html('<img id="imageManagerPreview" alt="image preview" src="' + fileSrc + '">');

      // Highlight row
      $('#fileList tr').removeClass('active');
      $('#' + fileId).addClass('active');
    }
  }

  /**
   * Init description and form after fileStorage input change
   */
  let storageChanged = function() {
    let storage = $('#fileStorage').val();
    let descriptionText = createDescriptionText(storage);
    let uploadForm = createUploadForm(storage);
    let form = descriptionText + uploadForm;

    options.onAfterStorageChanged();

    form = form.replace(/%ALLOWED_FILES%/g, options.messages.filesAllowed);
    $('#uploadForm').html(form);
  }


  /**
   * Convert storage type to information about that storage
   * @type {{ftp: {fieldName: string, tooltip: string, placeholder: string}, samba: {fieldName: string, tooltip: string, placeholder: string}, url: {fieldName: string, tooltip: string, placeholder: string}}}
   */
  const storageTypeInfo = {
    url: {
      fieldName: 'URL',
      tooltip: options.messages.urlTooltip,
      placeholder: 'http://' + app_name + '.com/image.png'
    },
    samba: {
      fieldName: 'Path',
      tooltip: options.messages.sambaTooltip,
      placeholder: '/example_path/image.png'
    },
    ftp: {
      fieldName: 'URL',
      tooltip: options.messages.ftpTooltip,
      placeholder: 'ftp://user:pass@' + app_name + '.com/image.png'
    },
  }

  /**
   * Create description text above form
   * @param {string} storageType
   * @returns {object}
   */
  let createDescriptionText = function(storageType) {
    const convertStorageTypeToDescText = {
      flash: options.messages.storageInfo.flash,
      SD0: options.messages.storageInfo.SD0,
      url: 'Set the URL to the image stored on the server and name it. The name will be used when selecting the displayed graphics. %ALLOWED_FILES%',
      samba: 'Connect your shared folder as the <a href="#" onclick="window.open(\'/#settings/system/services/storage\', \'_blank\')">camera\'s network storage</a> and enter the path to the saved image. Name this link for the subsequent selection of the graphics source for each service. %ALLOWED_FILES%',
      ftp: 'Set the URL to the image stored on your FTP server and name it. The name will be used when selecting the displayed graphics. %ALLOWED_FILES%'
    }
    return '<div class="mt-1 form-group row mb-1"><label class="col-lg-2"></label><small class="col-lg-10 text-form text-muted"><div class="text-left mb-1">' + convertStorageTypeToDescText[storageType] + '</div></small></div>';
  }

  /**
   * Create upload form
   * @param storageType
   * @returns {string}
   */
  let createUploadForm = function(storageType) {
    let form = '';
    if (storageUrlSambaFtp.includes(storageType)) {
      let storageInfo = storageTypeInfo[storageType];
      form +=  '<div id="urlNameFileGroup" class="form-group row mb-1" data-toggle="tooltip" title="This name will be shown in image list.">';
      form +=   '<label for="urlName" class="col-2 text-left control-label col-form-label">Name</label>';
      form +=   '<div class="col-10">';
      form +=     '<input type="text" class="form-control" id="urlName" placeholder="My Identifier">';
      form +=     '<span id="urlNameFileHelp" class="help-block"></span>';
      form +=   '</div>';
      form +=' </div>';
      form += '<div id="urlFileGroup" class="form-group row mb-1" data-toggle="tooltip" title="' + storageInfo.tooltip + '">';
      form +=   '<label for="url" class="col-2 text-left control-label col-form-label">' + storageInfo.fieldName + '</label>';
      form +=   '<div class="col-10">';
      form +=     '<input type="text" class="form-control" id="url" placeholder="' + storageInfo.placeholder + '">';
      form +=     '<span id="urlFileHelp" class="help-block"></span>';
      form +=   '</div>';
      form += '</div>';
    } else {
      form +=  '<div id="uploadedFileGroup" class="form-group row mb-1">';
      form +=  '<label for="filePath" class="col-2 text-left control-label col-form-label">File</label>';
      form +=   '<div class="col-10">';
      form +=     '<input class="form-control" type="file" multiple id="uploadedFile" name="uploadedFile[]" data-validation="required extension" data-validation-allowing="png,jpg">';
      form +=     '<small id="uploadFileHelper" class="mt-1 float-left text-form text-muted">Select one or more files from your computer.</small>';
      form +=     '<span id="uploadedFileHelp" class="help-block"></span>';
      form +=   '</div>';
      form += '</div>';
    }
    return form;
  }

  return {
    init,
    open,
    getFileList,
  };
}
