"use strict";

/**
 * Helper for js telemetry
 * - stores console output for diagnostic report
 * - provide handlers for sending fatal errors to server
 *
 * @constructor
 */
const JsLogger = function () {

  let buffers = {
    log: [],
    error: [],
    warn: [],
    uncaught_errors: [],
    unhandled_rejections: [],
  };

  let overload_console_methods = [ 'log', 'warn', 'error'];

  /**
   * @param item
   * @returns {*|string}
   */
  let getStackTraceFromItem = function (item)
  {
    return typeof item === 'object' && item.hasOwnProperty('stack') && typeof item.stack === 'string' ? item.stack : '';
  };

  /**
   * @returns {string}
   */
  let getActiveDomElement = function ()
  {
    let activeElement = document.activeElement.nodeName;
    if (document.activeElement.id) {
      activeElement += ", id: " + document.activeElement.id;
    }
    if (document.activeElement.name) {
      activeElement += ", name: " + document.activeElement.name;
    }
    if (document.activeElement.className) {
      activeElement += ", class: " + document.activeElement.className;
    }
    return activeElement;
  };

  let getLogs = function () {
    return buffers;
  };

  let onError = function (msg, url, lineNo, columnNo, error) {
    let error_data = {
      msg: msg,
      url: url,
      lineNo: lineNo,
      columnNo: columnNo,
      stack: getStackTraceFromItem(error),
    };

    let environment = {
      userAgent: window.navigator.userAgent,
      referrer: document.referrer,
      activeElement: getActiveDomElement(),
    };

    buffers.uncaught_errors.push({
      date: (new Date()).toUTCString(),
      ...error_data,
      environment,
    });
  };

  let onUnhandledRejection = function(e) {
    if (!e || !e.reason) {
      return false;
    }

    let reason = e.reason;
    let message;
    // try match jqXHR
    if (typeof reason === 'object' && reason.hasOwnProperty('statusText') && reason.hasOwnProperty('getResponseHeader')) {
      message = 'jqXHR: ' + reason.statusText + ' - ' + reason.getResponseHeader('x-tracy-error-log');
    } else {
      message = String(reason);
    }

    let data = {
      date: (new Date()).toUTCString(),
      reason: message,
      url: this.location.href,
      stack: getStackTraceFromItem(reason),
      environment: {
        userAgent: window.navigator.userAgent,
        referrer: document.referrer,
        activeElement: getActiveDomElement(),
      },
    };

    buffers.unhandled_rejections.push(data);
  };

  overload_console_methods.forEach(function(name) {
    let default_name = 'default' + name;
    console[default_name] = console[name].bind(console);
    console[name] = function () {
      console[default_name].apply(console, arguments);
      buffers[name].push(Array.from([(new Date()).toUTCString(), ...arguments]));
    }
  });

  return Object.freeze({
    getLogs,
    onError,
    onUnhandledRejection,
  });
};