const WS_RECONNECT_INTERVAL = 3000;

function SharingInterface() {

  const sharingInterfaceContainer = $('#sharingInterfaceContainer');
  let list = {};

  let getWsToken = async () => {
    try {
      let tokenResponse = await fetch('/local/camoverlay/api/ws_authorization.cgi');
      if (tokenResponse.status !== 200) {
        console.error("SharingInterface - Request failed (WS authorization token)");
        return null;
      }

      let tokenJson = await tokenResponse.json();
      if (tokenJson.status != 200 || !tokenJson.message) {
        console.error("SharingInterface - Invalid JSON response (WS authorization token)");
        return null;
      }
      return tokenJson.message;

    } catch (err) {
      console.error("SharingInterface - Can't get WS authorization token: " + err.message);
      return null;
    }
  }

  let event_code_to_msg = (event_code) => {
    switch (event_code) {
      case 1000: return 'Normal Closure';
      case 1001: return 'Going Away';
      case 1002: return 'Protocol error';
      case 1003: return 'Unsupported Data';
      case 1004: return 'Reserved';
      case 1005: return 'No Status Rcvd';
      case 1006: return 'Abnormal closure';
      case 1007: return 'Invalid frame payload data';
      case 1008: return 'Policy Violation';
      case 1009: return 'Message Too Big';
      case 1010: return 'Mandatory Ext.';
      case 1011: return 'Internal Server Error';
      case 1015: return 'TLS handshake';
      default: return "Code:" + event_code;
    }
  }

  let wsConnection = (token, url, name, service_id) => {
    return new Promise((resolve, reject) => {
      let ws = new WebSocket(url, name);
      ws.binaryType = 'blob';
      ws.onopen = () => {
        ws.send(JSON.stringify({
          authorization: token,
          serviceID: service_id
        }));
        $('#container_videoSource' + service_id).removeClass('reconnecting');
        resolve(ws);
      }
      ws.onclose = (event) => {
        if (existInList(service_id)) {
          $('#container_videoSource' + service_id).addClass('reconnecting');
          list[service_id].timeoutInterval = setTimeout(() => {
            wsReconnect(url, name, service_id);
          }, WS_RECONNECT_INTERVAL);
        } else {
          const msg = event_code_to_msg(event.code);
          reject(Error(msg));
        }
      }
    });
  }

  let wsReconnect = async (url, name, service_id) => {
    if (!existInList(service_id)) {
      return;
    }

    const token = await getWsToken();
    if (!token) {
      list[service_id].timeoutInterval = setTimeout(() => {
        wsReconnect(url, name, service_id);
      }, WS_RECONNECT_INTERVAL);
      return;
    }
    try {
      list[service_id].ws = await wsConnection(token, url, name, service_id);
    } catch (err) {
      console.log('SharingInterface - ws reconnect: ' + err.message, service_id)
    }
  }

  let wsIsOpened = (service_id) => {
    if (!existInList(service_id)) {
      return false;
    }
    return list[service_id].ws.readyState === WebSocket.OPEN;
  }

  /**
   * Create CANVAS with captured picture
   * @returns {HTMLCanvasElement}
   */
  let canvas_with_picture = (videoElement, capture, resolution, imageMaxSize, fullscreenOn) => {
    return new Promise((resolve, reject) => {
      const video = videoElement.get();
      const captureObject = capture.getSettings();
      const currentResolution = resolution;
      const currentImageMaxSize = imageMaxSize;
      const isFullscreenOn = fullscreenOn;

      const canvas = document.createElement("canvas");
      canvas.height = isFullscreenOn ? currentResolution.height : currentImageMaxSize;
      canvas.width = isFullscreenOn ? currentResolution.width : currentImageMaxSize;

      const ratio = Math.min(canvas.width / captureObject.width, canvas.height / captureObject.height);
      const new_width = captureObject.width * ratio;
      const new_height = captureObject.height * ratio;

      if (!isFullscreenOn) {
        if (new_height >= new_width) {
          canvas.width = new_width;
        } else {
          canvas.height = new_height;
        }
      }

      let offsetX = isFullscreenOn ? (canvas.width - new_width) / 2 : 0;
      let offsetY = isFullscreenOn ? (canvas.height - new_height) / 2 : 0;
      canvas.getContext('2d').drawImage(video, offsetX, offsetY, new_width, new_height);
      canvas.toBlob(resolve, 'image/jpeg');
    });
  }

  let wsSendInInterval = (service_id) => {
    if (!existInList(service_id)) {
      return;
    }

    // Stop already running worker
    if (list[service_id].sendingIntervalWorker) {
      list[service_id].sendingIntervalWorker.terminate();
    }

    // Start new worker
    list[service_id].sendingIntervalWorker = new Worker('js/WorkerInterval.js');
    list[service_id].sendingIntervalWorker.onmessage = async (evt) => {
      const imageData = await canvas_with_picture(
        list[service_id].videoElement,
        list[service_id].capture,
        list[service_id].resolution,
        list[service_id].imageMaxSize,
        list[service_id].fullscreenOn
      );
      wsSend(service_id, imageData);
    };
    list[service_id].sendingIntervalWorker.postMessage(list[service_id].fps);
  }

  let wsSend = (service_id, data) => {
    if (wsIsOpened(service_id)) {
      list[service_id].ws.send(data);
    }
  }

  let wsFullscreenOn = (service_id) => wsSend(service_id, JSON.stringify({command: "ENABLE_FULLSCREEN"}));
  let wsFullscreenOff = (service_id) => wsSend(service_id, JSON.stringify({command: "DISABLE_FULLSCREEN"}));
  let wsClose = (service_id) => list[service_id].ws.close();

  let containerShowOrHide = () => {
    Object.keys(list).length === 0 ? sharingInterfaceContainer.removeClass('open') : sharingInterfaceContainer.addClass('open');
  }

  let setResolution = (service_id, current_resolution) => {
    if (!existInList(service_id)) {
      return;
    }
    list[service_id].resolution = current_resolution;
  }

  let setFullscreen = (service_id, fullscreen) => {
    if (!existInList(service_id)) {
      return;
    }
    list[service_id].fullscreenOn = fullscreen;
    list[service_id].fullscreenOn ? wsFullscreenOn(service_id) : wsFullscreenOff(service_id);
  }
  let setImageMaxSize = (service_id, imageMaxSize) => {
    if (!existInList(service_id)) {
      return;
    }
    list[service_id].imageMaxSize = imageMaxSize;
  }

  let setFps = (service_id, frame_per_seconds) => {
    if (!existInList(service_id)) {
      return;
    }
    list[service_id].fps = frame_per_seconds;
    list[service_id].capture.setFps(frame_per_seconds);
    wsSendInInterval(service_id);
  }
  let getFullscreen = (service_id) => list[service_id].fullscreenOn;
  let getImageMaxSize = (service_id) => list[service_id].imageMaxSize;
  let getFps = (service_id) => list[service_id].fps;
  let isRunning = (service_id) => existInList(service_id);

  let start = async (type, ws_config, service_id, resolution, imageMaxSize, fullscreenOn, fps, onStopCallback) => {
    if (!window.Worker) {
      throw Error("Unsupported browser (Worker support is needed).");
    }

    const token = await getWsToken();
    if (!token) {
      throw Error("Can't get WS authorization token");
    }
    const websocket = await wsConnection(token, ws_config.url, ws_config.name, service_id);

    let capture = new Capture();
    let mediaStream;
    try {
      if (type === 'screen') {
        mediaStream = await capture.setScreen(fps);
      } else if (type === 'webcam') {
        mediaStream = await capture.setWebcam(fps);
      } else {
        throw Error('Unknown source');
      }
    } catch (err) {
      throw Error(err.message);
    }

    const videoElement = new VideoElement(service_id, sharingInterfaceContainer);
    await videoElement.create(mediaStream);

    videoElement.onLoadedData(() => {
      wsSendInInterval(service_id);
    });

    capture.onEnded(() => {
      stop(service_id);
    });

    const item = {
      service_id: service_id,
      resolution: resolution,
      imageMaxSize: imageMaxSize,
      fullscreenOn: fullscreenOn,
      fps: fps,
      ws: websocket,
      capture: capture,
      videoElement: videoElement,
      sendingIntervalWorker: null,
      timeoutInterval: null,
      onStop: onStopCallback,
    };
    list[service_id] = item;

    containerShowOrHide();
  }

  let stop = (service_id) => {
    if (!existInList(service_id)) {
      return;
    }

    if (list[service_id].sendingIntervalWorker) {
      list[service_id].sendingIntervalWorker.terminate();
    }
    list[service_id].capture.stop();
    list[service_id].videoElement.remove();
    list[service_id].onStop();
    wsClose(service_id);
    delete list[service_id];

    containerShowOrHide();
  }

  let existInList = (service_id) => {
    return list[service_id] !== undefined;
  }

  return {
    isRunning,
    start,
    stop,
    setResolution,
    setFullscreen,
    setImageMaxSize,
    setFps,
    getFullscreen,
    getImageMaxSize,
    getFps,
  }
}